"""
CIDR-Clustering Integration for annData using Rpy2

This module provides a Python function for the R-based CIDR algorithm for clustering
single-cell RNA-seq data using Rpy2.

Functions
cidr_rpy2: Memory-based CIDR clustering.
"""
import numpy as np

def cidr_rpy2(a_data, layer="X", data_type="raw", n_cluster=None, pc=False, save_clusters=True):
    """
    This function performs CIDR in R using an approach based on Python's Rpy2 package which is meant to interface
    with R through C scripts.

    Args:
        a_data (annData.AnnData): The anndata object containing expression data
        layer (str, optional): Name of the layer in 'a_data' to use. If "X", uses 'a_data.X'. Defaults to "X".
        data_type (str): Type of data ("raw" or "cpm")
        n_cluster (int or None): Number of clusters. If None, this is determined by the algorithm
        pc (bool, optional): Determines whether to save the principle coordinates, proportion of variation
            and eigenvalue vectors into the annData object.
            These are saved in annData.obsm, annData.uns and annData.uns respectively. Defaults to False.
        save_clusters (bool, optional): Determines whether to return clustering coordinates for different cells to the 
            annData object. These are saved into annData.obsm. Defaults to True.

    Returns:
        None
    """
    # Runs CIDR algorithm with rpy2.
    # The lines that start with ro.r is a way of running R-code in python.

    # Import of rpy2 is set here instead on top of module, making it possible to import the package
    # without needing to install rpy2, since this package might not always compile in a windows environment.

    try:
        import rpy2.robjects as ro
        from rpy2.robjects import pandas2ri
        from rpy2.robjects.packages import importr
        from rpy2.robjects import numpy2ri
        from rpy2.robjects.conversion import localconverter
        import rpy2.robjects as ro
    except ImportError:
        raise ImportError("rpy2 must be installed to use 'cidr_rpy2'")

    # Imports the CIDR package installed in R.
    cidr = importr("cidr")

    data = a_data.X if layer == "X" else a_data.layers[layer]
    dense_matrix = _to_dense_array(data)

    with localconverter(ro.default_converter + numpy2ri.converter):
        r_matrix = ro.conversion.py2rpy(dense_matrix)

    ro.globalenv['data_matrix'] = r_matrix
    ro.r('data_object <- as.matrix(data_matrix)')
    ro.globalenv['data_type'] = data_type
    ro.r('''if (data_type == "cpm"){
            cidr_obj <- scDataConstructor(t(data_object), tagType="cpm")
            } else {
            cidr_obj <- scDataConstructor(t(data_object))
            }''')
    print("Object created (1/8)")
    ro.r('cidr_obj <- determineDropoutCandidates(cidr_obj)')
    print("Determined dropout candidates (2/8)")
    ro.r('cidr_obj <- wThreshold(cidr_obj)')
    print("Determined threshold (3/8)")
    ro.r('cidr_obj <- scDissim(cidr_obj)')
    print("Created dissimilaity matrix (4/8)")
    ro.r('pdf("cidr_plots.pdf")')
    ro.r('cidr_obj <- scPCA(cidr_obj)')
    print("Finished principal component analysis (5/8)")
    ro.r('cidr_obj <- nPC(cidr_obj)')
    print("Determined number of principal components (6/8)")

    # If n_cluster, then it chooses n_cluster as amount of clusters, otherwise
    # letting the algorithm choose.
    if n_cluster == None:
        ro.r('cidr_obj <- scCluster(cidr_obj)')
    else:
        ro.globalenv['n_cluster'] = n_cluster
        ro.r('cidr_obj <- scCluster(cidr_obj, nCluster=as.integer(n_cluster))')

    print("Finished clustering (7/8)")
    ro.r('''
    plot(cidr_obj@PC[, c(1, 2)],
      col = cidr_obj@clusters,
      pch = cidr_obj@clusters,
      main = "CIDR Clustering",
      xlab = "PC1", ylab = "PC2"
    )
    dev.off()
    ''')
    print("Plot done (8/8)")

    # If save_clusters, then the clusters are saved in the a_data object.
    if save_clusters == True:
        clusters = ro.r('as.data.frame(cidr_obj@clusters)')
    
        from rpy2.robjects.conversion import localconverter
        from rpy2.robjects import default_converter

        with localconverter(default_converter + pandas2ri.converter):
            clusters_df = ro.conversion.rpy2py(clusters)

        a_data.obsm[layer + '_cidr_clusters'] = clusters_df.values

    # If pc true in cidr function, then principal coordinates it put in a_data object.
    if pc == True:
        pcs = ro.r('as.data.frame(cidr_obj@PC)')
        variation = ro.r('as.data.frame(cidr_obj@PC)')

        from rpy2.robjects.conversion import localconverter
        from rpy2.robjects import default_converter

        with localconverter(default_converter + pandas2ri.converter):
            pcs_df = ro.conversion.rpy2py(pcs)
            variation_df = ro.conversion.rpy2py(variation)

        # Chooses the layer in which you save the principal coordinates.

        a_data.obsm[layer + '_cidr_pca'] = pcs_df.values
        a_data.uns[layer + '_variation'] = variation_df

        print("Algorithm done. Plots are in cidr_plots.pdf")

# This function should handle most data types w.r.t the annData object
def _to_dense_array(data, dtype=np.float32):
    if hasattr(data, "toarray"):
        intermediate = data.toarray()
    else:
        intermediate = data

    dense_array = np.asarray(intermediate, dtype=dtype)
    return dense_array




import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from scipy.sparse import csr_matrix  # type: ignore
import pyarrow as py
import subprocess
import os
from io import StringIO

run_state = {}

def _sparse_to_csv_using_buffer(sparse_matrix):
    array_matrix = sparse_matrix.toarray()
    df = pd.DataFrame(array_matrix)
    in_ram = StringIO()
    df.to_csv(in_ram, index=False)
    return in_ram.getvalue()

def _args_all(func_name, adata_id, layer_name, **kwargs):
    # Make an argumentlist to the R-script."
    script_path = os.path.join(os.path.dirname(__file__), "module5.R")

    args = ["Rscript",
            script_path,
            str(func_name),
            str(adata_id),
            str(layer_name)]

    for k, v in kwargs.items():
        args.append(f"--{k}={v}")

    return args


def _get_state(adata_id, layer):
    return run_state.setdefault((adata_id, layer), {
        "scDataConstructor": False,
        "determineDropoutCandidates": False,
        "wThreshold": False,
        "scDissim": False,
        "scPCA": False,
        "nPC": False,
        "nCluster": False,
        "scCluster": False
    })

CIDR_DEPENDENCIES = {
    "scDataConstructor": [],
    "determineDropoutCandidates": ["scDataConstructor"],
    "wThreshold": ["determineDropoutCandidates"],
    "scDissim": ["wThreshold"],
    "scPCA": ["scDissim"],
    "nPC": ["scPCA"],
    "nCluster": ["nPC"],
    "scCluster": ["nCluster"]
}

DEFAULT_ARGS = {
    "scDataConstructor": {
        "tagType": "raw"},
    "determineDropoutCandidates": {
        "min1": 3, "min2": 8, "N": 2000, "alpha": 0.1, "fast": True, "zerosOnly": False, "bw_adjust": 1},
    "wThreshold": {
        "cutoff": 0.5, "plotTornado": False},
    "scDissim": {
        "correction": False, "threads": 0, "useStepFunction": True},
    "scPCA": {
        "plotPC": True},
    "nCluster": {
        "n": None, "nPC": None, "cMethod": "ward.D2"},
    "scCluster": {
        "n": None, "nCluster": None, "nPC": None, "cMethod": "ward.D2"}
}

def _run_cidr_step(step, adata, layer="X",force=False, **kwargs):
    adata_id = id(adata)
    state = _get_state(adata_id, layer)

    for dep in CIDR_DEPENDENCIES.get(step, []):
        _run_cidr_step(dep, adata, layer)

    # Skip if the prevoius layerwas ran and it isn't foricng a rerun
    if state[step] and not force:
        return

    step_args = DEFAULT_ARGS.get(step, {}).copy()
    step_args.update(kwargs)

    input_data = None

    if step == "scDataConstructor":
        data = adata.X if layer == "X" else adata.layers[layer]
        input_data = _sparse_to_csv_using_buffer(csr_matrix(data, dtype=np.float32)).encode("utf-8")

    subprocess.run(
        _args_all(step, adata_id, layer, **step_args),
        input = input_data,
        check = True,
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT
    )

    state[step] = True
    _save_to_adata(adata, layer, step)

def _save_to_adata(adata, layer, step):
    if "cidr" not in adata.uns:
        adata.uns["cidr"] = {}
    if layer not in adata.uns["cidr"]:
        adata.uns["cidr"][layer] = {"steps_run": [], "results":{}}

    step_results = {}

    for file_name in os.listdir():
        if file_name.endswith(".parquet"):
            key = file_name.replace(".parquet", "")
            try:
                df = pd.read_parquet(file_name)

                if df.shape[1] > 1:
                    value = df.values

                elif df.shape[1] == 1:
                    value = df.iloc[:, 0].tolist()

                else:
                    value = df

                step_results[key] = value

                os.remove(file_name)

            except Exception as e:
                print(f"Could not read {file_name}: {e}")

    adata.uns["cidr"][layer]["results"][step] = step_results

    adata.uns["cidr"][layer]["steps_run"].append(step)

def cidr_scDataConstructor(a_data, layer="X", **kwargs):
    _run_cidr_step("scDataConstructor", a_data, layer=layer, **kwargs)

def cidr_determineDropoutCandidates(a_data, layer="X", **kwargs):
    _run_cidr_step("determineDropoutCandidates", a_data, layer=layer, **kwargs)

def cidr_wThreshold(a_data, layer="X", **kwargs):
    _run_cidr_step("wThreshold", a_data, layer=layer, **kwargs)

def cidr_scDissim(a_data, layer="X", **kwargs):
    _run_cidr_step("scDissim", a_data, layer=layer, **kwargs)


def cidr_scPCA(a_data, layer="X", **kwargs):
    _run_cidr_step("scPCA", a_data, layer=layer, **kwargs)


def cidr_nPC(a_data, layer="X", **kwargs):
    _run_cidr_step("nPC", a_data, layer=layer, **kwargs)


def cidr_nCluster(a_data, layer="X", **kwargs):
    _run_cidr_step("nCluster", a_data, layer=layer, **kwargs)


def cidr_scCluster(a_data, layer="X", **kwargs):
    _run_cidr_step("scCluster", a_data, layer=layer, **kwargs)



"""
CIDR-Clustering Integration for annData

This module provides Python functions for the R-based CIDR algorithm for clustering
single-cell RNA-seq data. It supports both read/write file-based communication
between Python and R via subprocess calls.

Functions
cidr: File-based CIDR clustering
cidr_non_csv: In-memory CIDR clustering.
"""

import anndata as ad  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from scipy.sparse import csr_matrix  # type: ignore
import subprocess
import os
import tempfile
from io import StringIO

def _sparse_to_csv(sparse_matrix, csv_name):
    # Save a sparse matrix as a csv file
    array_matrix = sparse_matrix.toarray()
    df = pd.DataFrame(array_matrix)
    df.to_csv(csv_name, index=False)

def _sparse_to_csv_using_buffer(sparse_matrix):
    array_matrix = sparse_matrix.toarray()
    df = pd.DataFrame(array_matrix)
    # makes the buffer for an input stream
    in_ram = StringIO()
    df.to_csv(in_ram, index=False)
    return in_ram.getvalue()

def _make_args(data_type, n_cluster, pc, dropout, dissim, large, file_path=None, pc_path=None, var_path=None, eigen_path=None, dropout_path=None, dissim_path=None):
    # Make an argumentlist to the Rscript."
    script_path = os.path.join(os.path.dirname(__file__), "module2.R")

    return ["Rscript",
            script_path,
            str(data_type),
            str(n_cluster),
            str(pc),
            str(dropout),
            str(dissim),
            str(large),
            str(file_path),
            str(pc_path),
            str(var_path),
            str(eigen_path),
            str(dropout_path),
            str(dissim_path)]

def _handle_dropout(a_data, layer, dropout_src, buffer=False):
    read = lambda src: pd.read_csv(StringIO(src), index_col=0).to_numpy() if buffer else pd.read_csv(src, index_col=0).to_numpy()

    dropout_array = read(dropout_src)

    a_data.uns[layer + "_cidr_dropout_candidates"] = dropout_array


def _handle_dissim(a_data, layer, dissim_src, buffer=False):
    read = lambda src: pd.read_csv(StringIO(src), index_col=0).to_numpy() if buffer else pd.read_csv(src, index_col=0).to_numpy()

    dissim_array = read(dissim_src)

    a_data.obsp[layer + "_cidr_dissimilarity_matrix"] = dissim_array


def _handle_pc(a_data, layer, pc_src, var_src, eigen_src, buffer=False):
    read = lambda src: pd.read_csv(StringIO(src), index_col=0).to_numpy() if buffer else pd.read_csv(src, index_col=0).to_numpy()

    pc_array = read(pc_src)
    variation_array = read(var_src)
    eigen_array = read(eigen_src)

    a_data.obsm[layer + "_cidr_pc"] = pc_array
    a_data.uns[layer + "_cidr_variation"] = variation_array
    a_data.uns[layer + "_cidr_eigenvalues"] = eigen_array

def cidr(a_data, layer="X", data_type="raw", n_cluster=None, dropout=False, dissim=False, pc=False):
    """
    Perform CIDR on any layer of an annData object, using temporary read-write-files.
    This function is optimal if you are working with large datasets, that would otherwise
    hog a lot of RAM.

    .. warning::
        This algorithm uses temporary read-write files. These can be assumed deleted afterwards
        however, if a sudden spike in disc usage is experienced, these can be found and deleted in:
            For windows users:
            "C:\\Users\\your_user\\AppData\\Local\\Temp"
            For linux users:
            "/tmp"

    Args:
        adata (annData.AnnData): Input Anndata object containing expression data
        layer (str, optional): Name of the layer in 'a_data' to use. If "X", uses 'a_data.X'. Defaults to "X".
        data_type (str): Type of data ("raw" or "cpm")
        n_cluster (int or None): Number of clusters. If None, this is determined by the algorithm
        dropout (bool, optional): Determines whether to save dropout candidates into the annData object.
            These are saved in annData.uns. Defaults to False.
        dissim (bool, optional): Determines whether to save the dissimilarity matrix into the annData object.
            These are saved in annData.obsp. Defaults to False.
        pc (bool, optional): Determines whether to save the principle coordinates, proportion of variation
            and eigenvalue vectors into the annData object.
            These are saved in annData.obsm, annData.uns and annData.uns respectively. Defaults to False.

    Returns:
        None
    """

    # Run CIDR-clustering
    print("Algorithm Starting...")
    data = a_data.X if layer == "X" else a_data.layers[layer]
    sparse_matrix = csr_matrix(data, dtype=np.float32)

    # Creates randomly named temporary files. These are later deleted for the user.
    pc_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
    var_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
    eigen_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
    temp_data_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
    dropout_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
    dissim_file = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)

    # Sets path for different datafiles for use to read and write in R-script.
    data_path = temp_data_file.name
    pc_path = pc_file.name
    var_path = var_file.name
    eigen_path = eigen_file.name
    dropout_path = dropout_file.name
    dissim_path = dissim_file.name

    # Closes the files to save ram
    pc_file.close()
    var_file.close()
    eigen_file.close()
    temp_data_file.close()
    dropout_file.close()
    dissim_file.close() # garbage collection

    # Saves sparse matrix as csv and then runs the Rscript on the file.
    _sparse_to_csv(sparse_matrix, data_path)
    (subprocess.run(_make_args(data_type, n_cluster, pc, dropout, dissim,
                              True, data_path, pc_path, var_path,
                              eigen_path, dropout_path, dissim_path), check=True).stderr)

    if dropout:
        _handle_dropout(a_data, layer, dropout_path)

    if dissim:
        _handle_dissim(a_data, layer, dissim_path)

    # If pc true in cidr function, then principal coordinates is put in a_data object.
    if pc:
        _handle_pc(a_data, layer, pc_path, var_path, eigen_path)

    # Makes sure to remove the temp files
    for path in [pc_path, var_path, eigen_path, dropout_path, dissim_path, data_path]:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass

    print("Algorithm done. Plots are in cidr_plots.pdf")

def cidr_non_csv(a_data, layer="X", data_type="raw", n_cluster=None, dropout=False, dissim=False, pc=False):
    """
    Perform CIDR on any layer of an annData object, using in-memory RAM.
    This function is optimal if you are working with small datasets.

    .. warning::
        .. warning::
        For very large datasets this function can possibly time-out
        it is then recommended to use 'cidr' instead

    Args:
        adata (annData.AnnData): Input Anndata object containing expression data
        layer (str, optional): Name of the layer in 'a_data' to use. If "X", uses 'a_data.X'. Defaults to "X".
        data_type (str): Type of data ("raw" or "cpm")
        n_cluster (int or None): Number of clusters. If None, this is determined by the algorithm
        dropout (bool, optional): Determines whether to save dropout candidates into the annData object.
            These are saved in annData.uns. Defaults to False.
        dissim (bool, optional): Determines whether to save the dissimilarity matrix into the annData object.
            These are saved in annData.obsp. Defaults to False.
        pc (bool, optional): Determines whether to save the principle coordinates, proportion of variation
            and eigenvalue vectors into the annData object.
            These are saved in annData.obsm, annData.uns and annData.uns respectively. Defaults to False.

    Returns:
        None
    """
    # Run CIDR-clustering
    print("Algorithm Starting...")
    data = a_data.X if layer == "X" else a_data.layers[layer]
    sparse_matrix = csr_matrix(data, dtype=np.float32)
    input_data = _sparse_to_csv_using_buffer(sparse_matrix)
    # Runs the R-Script on the input_data
    result = subprocess.run(
        _make_args(data_type, n_cluster, pc, dropout, dissim, False),
        input=input_data.encode("utf-8"),
        check=True,
        stdout=subprocess.PIPE
    )

    output = result.stdout.decode("utf-8")

    if dropout:
        dropout_str, output = output.split("---END OF DROPOUT CANDIDATES---\n")
        _handle_dropout(a_data, layer, dropout_str, buffer=True)

    if dissim:
        dissim_str, output = output.split("---END OF DISSIMILARITY---\n")
        _handle_dissim(a_data, layer, dissim_str, buffer=True)

    # If pc true in cidr function, then principal coordinates it put in a_data object.
    if pc:
        pc_str, var_str = output.split("---END OF PC---\n")
        eigen_str, output = var_str.split("---END OF VARIATION---\n")
        _handle_pc(a_data, layer, pc_str, var_str, eigen_str, buffer=True)

    print("Algorithm done. Plots are in cidr_plots.pdf")



from .module1 import cidr
from .module3 import cidr_rpy2
from .module1 import cidr_non_csv

from .module4 import cidr_scDataConstructor
from .module4 import cidr_determineDropoutCandidates
from .module4 import cidr_wThreshold
from .module4 import cidr_scDissim
from .module4 import cidr_scPCA
from .module4 import cidr_nPC
from .module4 import cidr_nCluster
from .module4 import cidr_scCluster
